import os
from tqdm import tqdm
from argparse import Namespace
from ..utilities import *

class EbasDownloader:
   def __init__(self, loc=None, url=None, ftp=None, bs4_selector=None):
      """init

      Args:
          loc (str, optional): path to save raw data. Defaults to None.
          url (str, optional): url for ebas file list. Defaults to None.
          ftp (str, optional): url for ebas ftp server. Defaults to None.
          bs4_selector (str, optional): selector for bs4. Defaults to None.
      """
      self.loc, self.url, self.ftp, self.bs4_selector = loc, url, ftp, bs4_selector
      if loc is None:
         cwd = os.getcwd()
         self.loc = os.path.join(cwd, 'ebas', 'raw_data')
      else:
         self.loc = os.path.join(loc, 'raw_data')
         
      if not os.path.exists(self.loc):
         os.makedirs(self.loc)
         print(f"Make data folder {self.loc}...")
         
      if ftp is None:
         self.ftp = 'https://thredds.nilu.no/thredds/fileServer/ebas/'
      if url is None:
         self.url = 'https://thredds.nilu.no/thredds/catalog/ebas/catalog.html'
      if bs4_selector is None:
         self.bs4_selector = "tr > td > a > tt"
   
   def get_raw_files(self, conditions=None, download=False):
      """update local files using current ebas files.

      Args:
          conditions (dict, optional): files selection conditions. Defaults to None.
          download (bool, optional): whether performs download. Defaults to False.
      """
      loc_files = self.__get_loc_files()
      ftp_files = self.__get_ebas_files()
      
      # remove files no longer exists on ftp server
      del_files = list(filter(lambda x: x not in ftp_files, loc_files))
      print(f"{len(del_files)} files need to be deleted...")
      if len(del_files)>0:
         self.__del_files(del_files)
      
      if conditions is not None:
         if isinstance(conditions, dict):
            conditions = Namespace(**conditions)
         ftp_files = self.__select_files(ftp_files, conditions)
      new_files = list(filter(lambda x: x not in loc_files, ftp_files))
      
      print(f"{len(new_files)} files need to be downloaded...")
      
      if not download and len(new_files)>0:
         p = input("View file names? (y/n) ")
         if p.lower() in ['yes', 'y', 'Y', 'Yes']:
            if len(new_files)>0:
               print("Files will be downloaded:")
               for f in new_files:
                  print(f)
            if len(del_files)>0:
               print("Files will be deleted:")
               for f in del_files:
                  print(f)
         download = input("Start downloading? (y/n) ")
         download = True if download.lower() in ['yes', 'y', 'Y', 'Yes'] else False
         
      if download:
         self.__download_files(new_files) # download new files
         
   def __download_files(self, files):
      ftp_list = []
      for (_, f) in enumerate(files):
         ftp_list.append((self.ftp+"/"+f, self.loc))
      if len(ftp_list)>0:
         print("Start downloading files...")
         run_mp_async(ftp_get_file, ftp_list)
         print("\nDownload completed.")

   def __del_files(self, files):
      for f in files:
         os.remove(os.path.join(self.loc, f))
      print("Old files have been removed.")
     
   def __select_files(self, files, conditions):
      """select files with args from files list

      Args:
          files (list): online ebas file lists
          conditions (Namespace): selection conditions
          may include start_year, end_year, sites, matrix, componets
      """
      
      ma = []
      com = []
      
      selected = []
      for f in tqdm(files, desc="selecting ftp files..."):
         temp = f.split('.')
         site = temp[0]
         st = int(temp[1][0:4])
         ed = int(temp[2][0:4])
         matrix = temp[5]
         component = temp[4]
         
         ma.append(matrix)
         com.append(component)
         
         if conditions.start_year is not None:
            if int(conditions.start_year) > ed:
               continue
         if conditions.end_year is not None:
            if int(conditions.end_year) < st:
               continue
         if conditions.site is not None:
            if site not in conditions.site:
               continue
         if conditions.matrix is not None:
            if matrix not in conditions.matrix:
               continue
         if conditions.components is not None:
            if component not in conditions.components:
               continue
         
         selected.append(f)
      
      # print(sorted(set(ma)))
      # print(sorted(set(com)))
      
      return selected
   
   def __get_ebas_files(self):
      print("Requesting data from ebas sever...")
      files = bs4_get(self.url, self.bs4_selector)
      # the first line is "Ebas", not a file
      print(f"{len(files)-1} files found on ftp server.")
      return files[1:]
   
   def __get_loc_files(self):
      nc_files =[]
      files = os.listdir(self.loc)
      nc_files = list(filter(lambda x: x.endswith("nc"), files))      
      print(f"{len(nc_files)} raw data (*.nc) files have been downloaded.")
      
      return nc_files
 


import argparse
import numpy as np
import time
from .ebas_downloader import *
from .ebas_db import *


def main():
   description = """
-----------------------------------------------------------------------------------------------------------
Download EBAS data with specified arguments:
   1. starting year
   2. ending year
   3. site code
   4. matrix
   5. components
If no values are provide, all data will be downloaded.
The result will be exported as .csv file.

Available matrix:
'aerosol', 'aerosol_humidified', 'air', 'air+aerosol', 'air+pm10', 
'dried_moss', 'instrument', 'met', 'pm1', 'pm10', 'pm10_humidified', 
'pm10_non_volatile', 'pm10_pm1', 'pm10_pm25', 'pm1_humidified', 
'pm1_non_refractory', 'pm25', 'pm_eq110', 'pm_eq165', 'pm_eq25', 'pm_eq35', 
'pm_eq50', 'pm_eq75', 'precip', 'precip+dry_dep', 'precip_tot', 'wetdep'

Available components:
'1-3-butadiene', '1-butene', '1-methylnaphthalene', '1-methylphenanthrene', 
'2-2-4-trimethylpentane', '2-2-dimethylbutane', '2-2-dimethylpropane', '2-butenal', 
'2-methylbenzaldehyde', '2-methylbutane', '2-methylpentane', '2-methylpropane', 
'2-methylpropenal', '2-methylpropene', '2-oxopropanal', '2-propenal', 
'3-buten-2-one', '3-methylpentane', 'BDE_154', 'BDE_209', 'FTS_6-2', 'HCB', 
'NOx', 'NOy', 'PCB_114', 'PCB_118', 'PCB_150', 'PCB_153', 'PCB_18', 'PCB_180', 
'PCB_209', 'PCB_28', 'PCB_28+16', 'PCB_99', 'UVA_radiation', 'acenaphthene', 
'acenaphthylene', 'acidity', 'aerosol_absorption_coefficient', 'aerosol_light_backscattering_coefficient', 
'aerosol_light_scattering_coefficient', 'aerosol_optical_depth', 'aldrin', 'alpha_HCH', 
'aluminium', 'ammonia', 'ammonium', 'antimony', 'arsenic', 'barium', 'benz_a_anthracene', 
'benzaldehyde', 'benzene', 'benzo_a_pyrene', 'benzo_b_fluoranthene', 'benzo_bj_fluoranthenes', 
'benzo_bjk_fluoranthenes', 'benzo_k_fluoranthene', 'bromide', 'butanales', 'butanone', 
'cadmium', 'calcium', 'carbon_dioxide', 'carbon_monoxide', 'cesium', 'chloride', 'chromium', 
'chrysene', 'chrysene_triphenylene', 'cloud_condensation_nuclei_number_concentration', 
'cloud_condensation_nuclei_number_size_distribution', 'cobalt', 'conductivity', 'copper', 
'cyclo-hexane', 'dibenzo_ah_anthracene', 'dibenzofuran', 'dimethylsulfide', 'dinitrogen_monoxide', 
'elemental_carbon', 'endosulfan', 'equivalent_black_carbon', 'equivalent_black_carbon_mass', 
'erbium', 'ethanal', 'ethane', 'ethanedial', 'ethene', 'ethylbenzene', 'ethyne', 'fluorantene', 
'fluoride', 'gamma_HCH', 'gaseous_elemental_mercury', 'hexanal', 'hydrochloric_acid', 'hydrogen',
'hygroscopic_growth_factor', 'inden_123cd_pyrene', 'iron', 'lanthanum', 'lead', 'levoglucosan',
'm-p-xylene', 'magnesium', 'manganese', 'mercury', 'methanal', 'methane', 'methanesulfonic_acid', 
'molybdenum', 'n-butane', 'n-heptane', 'n-hexane', 'n-octane', 'n-pentane', 'nickel', 'niobium',
'nitrate', 'nitric_acid', 'nitrogen_dioxide', 'nitrogen_monoxide', 'o-xylene', 'organic_carbon',
'organic_carbon_corrected', 'organic_mass', 'ozone', 'pH', 'particle_number_concentration',
'particle_number_size_distribution', 'pentachlorobenzene', 'pentanal', 'pm10_mass', 'pm10_pm25_mass', 
'pm1_mass', 'pm25_mass', 'potassium', 'pp_DDD', 'pp_DDT', 'precipitation_amount', 
'precipitation_amount_off', 'pressure', 'propanal', 'propane', 'propanone', 'propene', 'propyne', 
'reactive_gaseous_mercury', 'relative_humidity', 'sample_count', 'scandium', 'sodium', 'strontium',
'sulphate_corrected', 'sulphate_total', 'sulphur_dioxide', 'sum_DDT', 'sum_PCB', 
'sum_ammonia_and_ammonium', 'sum_nitric_acid_and_nitrate', 'susp_part_matter', 'temperature', 'thallium', 
'tin', 'titanium', 'toluene', 'total_carbon', 'total_gaseous_mercury', 'trans-2-butene', 'vanadium', 
'voc_complete', 'wind_direction', 'wind_speed', 'zinc'

Important: If you are planning to download all the EBAS data 
           Do not use this command as the output csv file will be extremely large (>30GB).
           Do create EBAS database with pyebas.
           
Example: pyebas 2019 2021 --mode csv --site ES0010R ES0011R --matrix air --components NOx --output .\simple_csv_test
-----------------------------------------------------------------------------------------------------------
   """
   
   parser = argparse.ArgumentParser(prog="Simple download EBAS data",
                                    description=description,
                                    formatter_class=argparse.RawDescriptionHelpFormatter)
   
   parser.add_argument('start_year', type=int, help='starting year, eg. 1990', default=None)
   parser.add_argument('end_year', type=int, help='ending year, eg. 2021', default=None)
   parser.add_argument('--mode', choices=['db','csv', 'query'], help='export mode', default='csv')
   parser.add_argument('--site', nargs="*", help='site code', default=None)
   parser.add_argument('--matrix', nargs="*", help='matrix', default=None)
   parser.add_argument('--components', nargs="*", help='component names', default=None)
   parser.add_argument('--output', type=str, help='output path', default=None)
      
   args = parser.parse_args()
   
   downloader = EbasDownloader(args.output)
   
   if args.mode=="csv":
      # export to csv file
      downloader.get_raw_files(conditions=args)
      csv_exporter = csvExporter(args.output)
      csv_exporter.export_csv()
   
   elif args.mode=="db":
      # create local database
      downloader.get_raw_files(conditions=args, download=True)
      db = EbasDB(args.output, dump='xz', detailed=True)
      db.update_db()
      
   elif args.mode=="query":
      db = EbasDB(args.output, dump='xz', detailed=True)
      db.init_db()
      print("Query Mode (ctrl+c to exit)")
      while True:
         condition ={}
         print("Enter selecting conditions, seperated by space, empty means select all:")
         id = input("site id:")
         if id!="":
            condition["id"] = id.strip().split(' ')
         
         matrix = input("matrix:")
         if matrix!="":
            condition["matrix"] = matrix.strip().split(' ')

         component = input("component:")
         if component!="":
            condition["component"] = component.strip().split(' ')

         st = input("starting date:")
         if st!="":
            condition["st"] = np.datetime64(st.strip())      

         ed = input("ending date:")
         if ed!="":
            condition["ed"] = np.datetime64(st.strip())  
         
         print(condition)
         df = db.query(condition, use_number_indexing=False)
         if df is not None:
            print(df.head(20))
         else:
            print("No data found!")
         
         export = input("Export selected dataframe? (y/n)")
         
         if export in ["y","yes","Y","Yes","YES"]:
            print("Exporting...")
            t= time.strftime("%Y%m%d%H%M")
            out_file = os.path.join(args.output, f"pyebas_{t}.csv")
            df.to_csv(out_file, index=False)
            print(f"Data has been exported to {out_file}.")



if __name__ =="__main__":   
   main()
      


import os
import xarray as xr
from .site_index import *
from .value_index import *
from .query import *
from .files_io import *

class EbasDB(SiteIndex, ValueIndex):
   def __init__(self, dir=None, dump='xz', **agrs):
      """EBAS database

      Args:
          dir (str, optional): path to database. Defaults to None.
          dump (str, optional): dump file type, 'xz', 'p', or 'json'. Defaults to 'xz'.
                                note: json file can be only used for exporting, 
                                      it can't be used as indexing file in databas. 
      """
      self.dump = dump
      self.db_dir = None
      self.dump_dir = None
      self.raw_dir = None
      # index of sites, including site information, and file content
      self.site_index = None
      self.value_index = None
      # inverse indexing of file content in site_index
      self.db_index={}
      self.db={}
      self.selected = {}
            
      self.bad_qc = [459,460,471,530,533,540,549,565,566,567,568,591,599,635,658,659,663,664,666,669,677,682,683,684,685,686,687,699,783,890,980,999]
      self.__make_dir(dir)
      super(EbasDB, self).__init__(raw_dir=self.raw_dir, 
                                   dump_dir=self.dump_dir,
                                   dump=dump,
                                   **agrs)
      
   def init_db(self):
      print("init database...")     
      print("Load value index...")
      file_path = os.path.join(self.db_dir, f"value_index.{self.dump}")
      self.value_index = load_value(file_path) 
      
      print("Load site index...")
      file_path = os.path.join(self.db_dir, f"site_index.{self.dump}")
      self.site_index = load_value(file_path)
      
      print("Load site data...")    
      files = []
      suffix = '.xz' if self.dump=="xz" else '.p'
      for site in self.site_index.keys():
         temp = {"path": os.path.join(self.dump_dir, site+suffix),
               "name": site}
         files.append(temp)
      self.db_index, self.db = load_files(files)
      Query.db_overview(self.site_index)
      print("Database is loaded.")
      
   def update_db(self):
      # create site index
      self.create_site_index()
      # create value index
      self.create_value_index(self.site_index)
      dump_value(self.value_index, self.db_dir, "value_index", self.dump)
      # update site index with value index
      self.site_index = self.update_site_index(self.site_index)
      dump_value(self.site_index, self.db_dir, "site_index", self.dump)
      # import data
      self.__import_data()
   
   
   def query(self, query_dict, use_number_indexing=True):
      query_res = Query.query(self.site_index, self.db_index, query_dict, self.value_index)
      df = Query.get_df(self.db, self.db_index, query_res, use_number_indexing, self.value_index)
      return df
   
   def __import_data(self):
      arg_list = []
      for k in self.site_index.keys():
         arg_list.append((self.raw_dir, self.dump_dir, self.dump, self.site_index[k]["files"]))
      
      print("Importing datafile for each site...")
      utilities.run_mp(import_site_data, arg_list)


   def __make_dir(self, dir):
      if dir is None:
         cwd = os.getcwd()
         self.db_dir = os.path.join(cwd, 'ebas')
         self.dump_dir = os.path.join(cwd, 'ebas', 'dumps')
         self.raw_dir = os.path.join(cwd, 'ebas', 'raw_data')
      else:
         self.db_dir = dir
         self.dump_dir = os.path.join(dir, 'dumps')
         self.raw_dir = os.path.join(dir, 'raw_data')
         
      if not os.path.exists(self.db_dir):
         os.makedirs(self.db_dir)
         print(f"Make data folder {self.db_dir}...")
      if not os.path.exists(self.dump_dir):
         os.makedirs(self.dump_dir)
         print(f"Make data folder {self.dump_dir}...")
      if not os.path.exists(self.raw_dir):
         os.makedirs(self.raw_dir)
         print(f"Make data folder {self.raw_dir}...")



from .files_io import *
from .query import *

class ValueIndex:
   def __init__(self):      
      # bad qc values from EBAS website      
      self.meta={"ebas":0, "no_ebas":1, 0:"ebas", 1:"no_ebas"}
      self.matrix={}
      self.unit={}
      self.res_code={}
      self.component={}
      self.site={}
   
   def create_value_index(self, site_index):
      print("creating value index...")
      matrix = Query.summary_attr(site_index, "matrix")
      unit = Query.summary_attr(site_index, "unit")
      res_code = Query.summary_attr(site_index, "res_code")
      component = Query.summary_attr(site_index, "component")
      site = list(site_index.keys())
      
      self.update_value_index("matrix", matrix)
      self.update_value_index("unit", unit)
      self.update_value_index("res_code", res_code)
      self.update_value_index("component", component)
      self.update_value_index("site", site)
      
      self.value_index = {
         "matrix":self.matrix,
         "unit":self.unit,
         "res_code":self.res_code,
         "component":self.component,
         "site":self.site,
         "meta": self.meta
      }
      
   def update_site_index(self, site_index):
      """
      update site index with value index (convert string to number)
      """
      for site in site_index.keys():
         for f in site_index[site]["files"].keys():
            for index, content in enumerate(site_index[site]["files"][f]["contents"]):
               site_index[site]["files"][f]["contents"][index]["site"] = self.value2index("site", content["site"])
               site_index[site]["files"][f]["contents"][index]["matrix"] = self.value2index("matrix", content["matrix"])
               site_index[site]["files"][f]["contents"][index]["component"] = self.value2index("component", content["component"])
               site_index[site]["files"][f]["contents"][index]["unit"]= self.value2index("unit", content["unit"])
               site_index[site]["files"][f]["contents"][index]["res_code"] = self.value2index("res_code", content["res_code"])
               site_index[site]["files"][f]["contents"][index]["meta"] = self.value2index("meta", content["meta"])

      return site_index
   
   def update_value_index(self, attr_name, vals):
      temp = {}
      for index, val in enumerate(vals):
         temp[index]= val
         temp[val]= index
      
      setattr(self, attr_name, temp)
      
   def value2index(self, attr, val):
      return self.value_index[attr][val]


from ..utilities import *
import pickle
import lzma
import json
import os
import numpy as np
import xarray as xr

bad_qc = [459,460,471,530,533,540,549,565,566,567,568,591,599,635,658,659,663,664,666,669,677,682,683,684,685,686,687,699,783,890,980,999]


def load_files(files):
   """this method opens ebas data files
   
   Args:
         files (dict): {"path":"", "name":""}   
   Returns:
         (tuple): (db_index, db)    
   """
   db_index = {}
   db = {}
   res = run_mp(load_file, files)
         
   # combine all the data
   for r in res:
      db_index[r["name"]] = r["data"]["content_index"]
      r["data"].pop("content_index")
      db[r["name"]] = r["data"]

   
   return db_index, db

def load_file(file):
   """this method opens one '.xz', '.json', and python pickle files

   Args:
         file (dict): {"name":"", "path":""}

   Returns:
         dict: {"name":"", "data":""}
   """
   
   file_path = file["path"]
   
   if file_path.endswith("xz"):
      with lzma.open(file_path, "rb") as pickle_file:
         res = pickle.load(pickle_file)
   elif file_path.endswith("json"):
      with open(file_path,"r") as json_file:
         res = json.load(json_file)
   else:
      with open(file_path, "rb") as pickle_file:
         res = pickle.load(pickle_file)
   
   return {"name":file["name"], "data":res}

def load_value(file_path):
   if file_path.endswith("xz"):
      with lzma.open(file_path, "rb") as pickle_file:
         res = pickle.load(pickle_file)
   elif file_path.endswith("json"):
      with open(file_path,"r") as json_file:
         res = json.load(json_file)
   else:
      with open(file_path, "rb") as pickle_file:
         res = pickle.load(pickle_file)
   
   return res

def dump_value(var, dir, file_name, dump): 
   f_name = f"{file_name}.{dump}"
   print(f"Dumping data to to '{f_name}'...")      
   if dump =="xz":
      with lzma.open(os.path.join(dir, f_name), "wb") as pickle_file:
         pickle.dump(var, pickle_file)
   elif dump=="p":
      with open(os.path.join(dir, f_name),"wb") as f:
         pickle.dump(var, f)  
   else:
      with open(os.path.join(dir, f_name),"w") as f:
         json.dump(var, f, indent=4,  sort_keys=True, default=str)
         
def import_site_data(args):
   """
   {
      content_index: {
         id1:{
            st, ed, comp, var, matrix, res_code, units
         } ,
         id2:{
            }
      ,...
      }
      id1:(df: st, ed, val_qc)
      id2:(df: st, ed, val_qc)
   }
   """
   raw_dir, dump_dir, dump, files = args
   res = { "content_index" :{} }
   id_count = 0
   for file in files.keys():
      try:
         site_id = file.split(".")[0]
         ebas = xr.open_dataset(os.path.join(raw_dir, file))
         for content in files[file]["contents"]:
            res["content_index"][id_count] = {
                  "st": content["st"],
                  "ed": content["ed"],
                  "component":content["component"],
                  "matrix": content["matrix"],
                  "res_code": content["res_code"],
                  "unit": content["unit"],
                  "var": content["var"],
                  "stat": content["stat"],
                  "file":file
            }
            # temp = pd.DataFrame()
            st = ebas["time_bnds"].data[:,0]
            ed = ebas["time_bnds"].data[:,1]
            ts = np.array([st,ed]).T
            
            # get var value and qc, the values can be updated for several times, so additional dimensions may be applied
            val = ebas[content["var"]].data
            while len(val.shape)>1:
               val = val[-1,:]  
            qc = ebas[content["var"]+"_qc"].data
            while len(qc.shape)>1:
               qc = qc[-1,:]
            
            # filter value with qc values
            val[np.isin(qc, bad_qc)]=None
            
            val = np.array([val]).T
            
            res[id_count] = {"ts": ts, "val": val}
            id_count+=1
         
      except Exception as e:
         print(e, file)  
   
   if dump=="xz":   
      with lzma.open(os.path.join(dump_dir, f"{site_id}.xz"), "wb") as pickle_file:
         pickle.dump(res, pickle_file)
   else:
      with open(os.path.join(dump_dir, f"{site_id}.p"), "wb") as pickle_file:
         pickle.dump(res, pickle_file)


import numpy as np
import pandas as pd
import itertools
from tqdm import tqdm

class Query:
   @staticmethod
   def db_overview(site_index):
      country = list(set(map(lambda x: site_index[x]["country"], site_index)))
      matrix = list(map(lambda x: list(site_index[x]["matrix"].keys()), site_index))
      matrix = list(set(itertools.chain(*matrix)))

      components = list(map(lambda x: list(site_index[x]["components"].keys()), site_index))
      components = list(set(itertools.chain(*matrix)))
           
      print(f"{len(site_index.keys()):<10} sites included in current database.")
      print(f"{len(components):<10} components included in current database.")
      print(f"{len(matrix):<10} matrix included in current database.")
      print(f"{len(country):<10} country included in current database.")
      
   # @staticmethod
   # def summary(site_index, site_keys=None):
   #    # generate summary for all sites or selected sites
   #    if site_keys is None:
   #       site_keys = site_index.keys()
         
   #    res = {}      
   #    for k in site_keys:
   #       res["site"]={
   #          "country":site_index[k]["country"],
   #          "land_use":site_index[k]["land_use"],
   #          "station_setting":site_index[k]["station_setting"],
   #          "name":site_index[k]["name"],
   #          "id":site_index[k]["id"],
   #          "lat":site_index[k]["lat"],
   #          "lon":site_index[k]["lon"],
   #          "alt":site_index[k]["alt"],
   #          "components": Query.summary_attr(site_index, "component", site_keys=[k]),
   #          "matrix": Query.summary_attr(site_index, "matrix", site_keys=[k]),
   #          "st": Query.summary_attr(site_index, "st", site_keys=[k]),
   #          "ed": Query.summary_attr(site_index, "ed", site_keys=[k]),
   #          }
      
   #    return res
   
   @staticmethod
   def summary_attr(site_index, attr, site_keys=None):
      # attr can be anything in contents: matrix, res_code, unit, var, component
      if site_keys is None:
         site_keys = site_index.keys()
      res = []
      for site in site_keys:
         files = site_index[site]["files"]
         for file in files.keys():
            contents = files[file]["contents"]
            for c in contents:
               res.append(c[attr])
      res = list(set(res))
      res.sort()
      return res
   
   @staticmethod
   def query(site_index, db_index, condition, value_index):
      selected = {}
      condition_key = list(condition.keys())
      site_selector_key = ["id", "name", "land_use", "station_setting","country"]
      site_selector_key = list(set(site_selector_key) & set(condition_key))
      # if "id" in site_selector_key:
      #    condition["id"] = list(map(lambda x: value_index["site"][x], condition["id"]))
            
      measure_selector_key = ["component", "matrix","stat"]
      measure_selector_key = list(set(measure_selector_key) & set(condition_key))
      
      if "component" in measure_selector_key:
         cond_temp = []
         for c in condition["component"]:
            if c in value_index["component"].keys():
               cond_temp.append(value_index["component"][c])
         condition["component"] = cond_temp
         # condition["component"] = list(map(lambda x: value_index["component"][x], condition["component"]))
      if "matrix" in measure_selector_key:
         cond_temp = []
         for c in condition["matrix"]:
            if c in value_index["matrix"].keys():
               cond_temp.append(value_index["matrix"][c])
         condition["matrix"] = cond_temp
         # condition["matrix"] = list(map(lambda x: value_index["matrix"][x], condition["matrix"]))
  

      time_selector_key =  ["st", "ed"]
      time_selector_key = list(set(time_selector_key) & set(condition_key))
      
      time_selector = {}
      for k in time_selector_key:
         time_selector[k] = condition[k]
      for site_id in tqdm(site_index.keys(), desc="seraching..."):
         select =True
         if len(site_selector_key)>0:
            for key in site_selector_key:
               if site_index[site_id][key] not in condition[key]:
                  select=False
                  break
               
         # if the site is not the target, examing files are not necessary.
         if not select:
            continue
         
         files = []
         for file_id in db_index[site_id].keys():
            select_file = True
            content = db_index[site_id][file_id]
            if "st" in site_selector_key:
               if content["ed"]<condition["st"]:
                  select_file = False
            if "ed" in site_selector_key:
               if content["st"]>condition["ed"]:
                  select_file = False
            for k in measure_selector_key:
               if content[k] not in condition[k]:
                  select_file = False
                  break
            if select_file:
               files.append(file_id)
         
         if len(files)>0:
            selected[site_id] = files
      
      return (selected, time_selector)
   
   @staticmethod
   def get_df(db, db_index, query_res, use_number_index, value_index):
      print("Gathering data to dataframe...")
      selected, time_selector = query_res
      res_ts = []
      res_val = []
      
      for site in tqdm(selected.keys()):
         site_id = site
         for file in selected[site_id]:
            header = db_index[site_id][file]
            ts = db[site_id][file]["ts"]
            val = db[site_id][file]["val"]
            
            if len(time_selector)>0:
               if "st" in time_selector.keys():
                  st = time_selector["st"]
                  index = ts[:,0]>= st
                  ts = ts[index]
                  val = val[index]
               if "ed" in time_selector.keys():
                  ed = time_selector["ed"]
                  index = ts[:,1]<=ed
                  ts = ts[index]
                  val = val[index]           
            
            infor = np.empty((ts.shape[0], 4))
            infor[:,0] = value_index["site"][site_id]
            infor[:,1] = header["component"]
            infor[:,2] = header["unit"]
            infor[:,3] = header["matrix"]
                        
            val = np.hstack((val,infor))
            
            res_ts.append(ts)
            res_val.append(val)
      
      df = None
      if len(res_ts)>0:
         res_ts = np.vstack(tuple(res_ts))
         res_val = np.vstack(tuple(res_val))
         
         df1 = pd.DataFrame(res_ts, columns=["st", "ed"])
         df2 = pd.DataFrame(res_val, columns=["val", "site","component","unit","matrix"])
         df = pd.concat([df1,df2],axis=1)

         if not use_number_index:
            df["site"] =df["site"].map(value_index["site"])
            df["component"] =df["component"].map(value_index["component"])
            df["unit"] =df["unit"].map(value_index["unit"])
            df["matrix"] =df["matrix"].map(value_index["matrix"])
         
      return df


import os
import xarray as xr
from tqdm import tqdm
import numpy as np
import pandas as pd
import time
import json

class csvExporter:
   def __init__(self, loc=None):
      if loc is None:
         cwd = os.getcwd()
         self.loc = os.path.join(cwd, 'ebas','raw_data')
         self.csv_loc = os.path.join(cwd, 'ebas')
      else:
         self.loc = os.path.join(loc,'raw_data')
         self.csv_loc = loc
      if not os.path.exists(self.loc):
         os.makedirs(self.loc)
         print(f"Make data folder {self.loc}...") 
         
   def export_csv(self, file_name=None):
      res = pd.DataFrame()
      
      files = os.listdir(self.loc)
      nc_files = list(filter(lambda x: x.endswith("nc"), files))
      for file in tqdm(nc_files, desc="Processing files..."):
         ebas = xr.open_dataset(os.path.join(self.loc, file))
         ebas_metadata = ebas.ebas_metadata
         ebas_metadata = json.loads(ebas_metadata)
         
         # file time series
         st = np.array(ebas["time_bnds"].data[:,0])
         ed = np.array(ebas["time_bnds"].data[:,1])
         
         # file site information
         site_id = ebas_metadata["Station code"]
         # site_name = ebas_metadata["Station name"]
         try:
            alt = ebas_metadata["Station altitude"]                           
         except:
            alt = ""
         try:           
            lat = ebas_metadata["Station latitude"]                           
         except:
            lat = ""
         try:            
            lon = ebas_metadata["Station longitude"]               
         except:
            lon = ""      
         
         # file variable information
         vars = list(ebas.data_vars.keys())
         vars = list(filter(lambda x: not x.endswith("_qc") and not x.endswith("_ebasmetadata"), vars))
         vars.remove("time_bnds")
         vars.remove("metadata_time_bnds")   
         for v in vars:
            temp = ebas[v+"_ebasmetadata"].data.tolist()[-1]
            while isinstance(temp, list):
               temp = temp[-1]
            temp = json.loads(temp)
            
            if "Matrix" in temp.keys():
               matrix = temp["Matrix"],
               unit = temp["Unit"],
               stat = temp["Statistics"],
               component = temp["Component"],
            elif "ebas_matrix" in temp.keys():
               matrix = temp["ebas_matrix"],
               unit = temp["ebas_unit"],
               stat = temp["ebas_statistics"],
               component = temp["ebas_component"],
            
            # get var value and qc, the values can be updated for several times, so additional dimensions may be applied       
            val = ebas[v].data
            while len(val.shape)>1:
               val = val[-1,:]  
            qc = ebas[v+"_qc"].data
            while len(qc.shape)>1:
               qc = qc[-1,:]
            
            temp_pd = pd.DataFrame({"st":st, "ed":ed})
            
            temp_pd["site"] = site_id
            # temp_pd["name"] = site_name
            temp_pd["alt"] = alt
            temp_pd["lat"] = lat
            temp_pd["lon"] = lon
            temp_pd["val"] = val
            temp_pd["qc"] = qc
            temp_pd["unit"] = unit[0]
            temp_pd["matrix"] = matrix[0]
            temp_pd["stat"] = stat[0]
            temp_pd["component"] = component[0]    
            
            res = pd.concat([res,temp_pd], ignore_index= True)   
            
      print("Exporting...")
      if file_name is None:
         t= time.strftime("%Y%m%d%H%M")
         file_name = f"pyebas_{t}.csv"
      out_file = os.path.join(self.csv_loc, file_name)
      res.to_csv(out_file, index=False)
      print(f"Data has been exported to {out_file}.")
   
   


import concurrent.futures
import multiprocessing
from tqdm import tqdm
import csv
import pycountry
import wget
import os


from bs4 import BeautifulSoup
import requests



# simple web scrap for sth
def bs4_get(url, selector, tags=None):
   # tags = none, will return a list of target text
   # tags= [], will return a list of dict [{tag1:content, tag2: content}, ...]
   wb_res = requests.get(url)
   if wb_res.status_code !=200:
      raise ValueError("Connection error.")

   soup = BeautifulSoup(wb_res.text, 'lxml')
   
   selected = soup.select(selector)
   res = []
   for i in selected:
      if tags is None:
         res.append(i.text)
      else:
         temp={}
         for t in tags:
            temp[t] = i.get(t)
         
         res.append(temp)
      
   return res

def ftp_get_file(ftp, out_path):
   try:
      wget.download(ftp, out=out_path)
   except Exception as e:
      print(e)
      with open(os.path.join(out_path, 'ftp_error.txt'),'a') as f:
         f.write(ftp+'\n')

def run_mp_async(map_func, arg_list,num_cores=None):
   if num_cores is None:
      num_cores = multiprocessing.cpu_count()
      num_cores = len(arg_list) if len(arg_list)<num_cores else num_cores
   print(f"Using {num_cores} threads...")
   
   pool = multiprocessing.Pool(num_cores)
   for arg in arg_list:
      pool.apply_async(map_func,  args=arg) 
   pool.close()
   pool.join()

def run_mp(map_func, arg_list, combine_func=None, num_cores=None):
   if num_cores is None:
      num_cores = multiprocessing.cpu_count()
      num_cores = len(arg_list) if len(arg_list)<num_cores else num_cores
   print(f"Using {num_cores} threads...")
   
   with concurrent.futures.ProcessPoolExecutor(max_workers=num_cores) as pool:
      with tqdm(total=len(arg_list)) as progress:
         futures = []
         for args in arg_list:
               future = pool.submit(map_func, args)
               future.add_done_callback(lambda p: progress.update())
               futures.append(future)

         results = []
         
         for future in futures:
               result = future.result()
               results.append(result)
   
   if combine_func is not None:
      return combine_func(results)
   else:
      return results   
         

def list2csv(data, file_name, header=None, single_col=True):
   with open(file_name, 'w', newline='', encoding="utf-8") as f:
    write = csv.writer(f)
    if header is not None:
      write.writerow(header)
    if single_col:
      for item in data:
          write.writerow([item])
    else:
       write.writerows(data)
   
   print(f"Data is written to {file_name}.")
   

# convert country name and code 
def country2code(country):
   return pycountry.countries.search_fuzzy(country)[0].alpha_2
def code2country(country_code):
   return pycountry.countries.get(alpha_2=country_code).name





from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.1.4'
DESCRIPTION = 'Easy download and export EBAS data'
LONG_DESCRIPTION = 'Python package for an easy-access to open-source air pollutant data from EBAS database via FTP links.'


# Setting up
setup(
    name="pyebas",
    version=VERSION,
    author="Chuanlong Zhou",
    author_email="<zhouchuanlong1988@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    
    install_requires=["numpy","pandas","xarray","pycountry","bs4","wget","tqdm"],
    url = 'https://github.com/defve1988/pyebas',
    entry_points={
        'console_scripts': [
            'pyebas = pyebas.main:main'
        ]
    },
    keywords=['python', 'ebas', 'air pollution', 'ftp'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import os 
import xarray as xr
import json
import numpy as np

from ..utilities import *
from .files_io import *


class SiteIndex():
   def __init__(self, raw_dir, dump_dir, dump='xz', detailed=True):
      """site index

      Args:
          raw_dir (str): raw data path
          detailed (bool, optional): whether save detail descriptions from .nc file to database. Defaults to True.

      SiteIndex example:
         {
           id: site id,
           name: site name,
           country: Armenia,
           land_use: land use type,
           station_setting: station type,
           alt: 2080.0 m,
           lat: 40.38444444,
           lon: 44.260583333,
           components:{
              <component name>:{
                 st: starting date, datetime,
                 ed: ending date, datetime
              }
           }
           files: {
              <file_name>:{
                 content:[
                    <dict of each component in this file>, eg:  
                        {
                           res_code: resolution,
                           matrix: matrix,
                           unit: unit,
                           component: component name,
                           stat: statistics type,
                           st: starting date, datetime,
                           ed: ending date, datetime,
                           site: site id,
                           var: var name in nc file,
                        }
                 ],
              }
           }     
          }
      """
      self.raw_dir = raw_dir
      self.dump = dump
      self.detailed = detailed
      self.site_index = None
      self.dump_dir = dump_dir
      
      super(SiteIndex, self).__init__()
   
   def create_site_index(self):
      print("Gathering site information...")
      # filter out non ".nc" files
      raw_data = os.listdir(self.raw_dir)
      files = list(filter(lambda x: x.endswith('nc'), raw_data))
      self.site_index = run_mp(self.get_file_indexing, files, self.combine_file_index)
      
      print(f"Collected site number: {len(self.site_index.keys())}")
      # analysis bad files
      bad =[]
      for k in self.site_index.keys():
         if k.startswith(self.raw_dir):
            bad.append(k)

      if len(bad)>0:
         print(f"Bad files :{len(bad)}")
         for b in bad:
            print(b)
      else:
         print("No bad files were found.")
      
   def get_file_indexing(self, file_name):
      """gathering indexing information from one .nc file

      Args:
          file_name (str): .nc file name

      Returns:
          dict: indexing information, similar as SiteIndex
      """
      try:
         # get site information
         ebas = xr.open_dataset(os.path.join(self.raw_dir, file_name))
         ebas_metadata = ebas.ebas_metadata
         ebas_metadata = json.loads(ebas_metadata)
            
         site = {
                "id": ebas_metadata["Station code"],
                "name": ebas_metadata["Station name"],
                "country": code2country(ebas_metadata["Station code"][0:2]),
                "land_use":None,
                "station_setting":None,
                "alt":None,
                "lat":None,
                "lon":None,
                "files":{},
               #  "var_content":[]
            }
         try:
            site["land_use"] = ebas_metadata["Station land use"]           
         except:
            pass
         try:
            site["station_setting"] = ebas_metadata["Station setting"]         
         except:
            pass
         try:
            site["alt"] = ebas_metadata["Station altitude"]                           
         except:
            pass
         try:           
            site["lat"] = ebas_metadata["Station latitude"]                           
         except:
            pass
         try:            
            site["lon"] = ebas_metadata["Station longitude"]               
         except:
            pass
         
         # get var content
         vars = list(ebas.data_vars.keys())
         vars = list(filter(lambda x: not x.endswith("_qc") and not x.endswith("_ebasmetadata"), vars))
         vars.remove("time_bnds")
         vars.remove("metadata_time_bnds")
         
         var_content = []
         for v in vars:
            temp = ebas[v+"_ebasmetadata"].data.tolist()[-1]
            
            while isinstance(temp, list):
               temp = temp[-1]
               
            temp = json.loads(temp)
            
            if "Matrix" in temp.keys():
               content ={
                  "res_code": ebas_metadata["Resolution code"],
                  "matrix": temp["Matrix"],
                  "unit":temp["Unit"],
                  "meta": "no_ebas", 
                  
                  "var":v,
                  "site": ebas_metadata["Station code"],
                  "stat": temp["Statistics"],
                  "component":temp["Component"],
                  "st":ebas["time_bnds"][0,0].values,
                  "ed":ebas["time_bnds"][-1,1].values,
               }
            elif "ebas_matrix" in temp.keys():
               content ={
                  # "res": res_code_index[ebas_metadata["Resolution code"]],
                  # "matrix": matrix_index[temp["ebas_matrix"]],
                  # "unit":units_index[temp["ebas_unit"]],
                  # "meta": meta_index["no_ebas"],
                  
                  "res_code": ebas_metadata["Resolution code"],
                  "matrix": temp["ebas_matrix"],
                  "unit":temp["ebas_unit"],
                  "meta": "no_ebas",
                  
                  "var":v,
                  "site": ebas_metadata["Station code"],
                  "stat": temp["ebas_statistics"],
                  "component":temp["ebas_component"],
                  "st":ebas["time_bnds"][0,0].values,
                  "ed":ebas["time_bnds"][-1,1].values,
               }
               
            var_content.append(content)
            
         # get attr information
         attr_content={}
         if self.detailed:
            attrs = ebas.attrs
            attr_content ={}
            for a in attrs:
               temp = getattr(ebas,a)
               if isinstance(temp, np.ndarray):
                  temp= temp.tolist()
               attr_content[a] = temp
         
         site["files"] = {file_name:{"contents": var_content, "detail_attrs": attr_content}}
                       
         return {site["id"]: site}
      
      except Exception as e:
         print(e)
         print(file_name)
         return {file_name: {
                "id": "",
                "name": "",
                "land_use": "",
                "station_setting": "",
                "lat": "",
                "lon": "",
                "alt": "",
                "error":str(e),
            }}
         
   def combine_file_index(self, list_dict_res):
      res = {}
      for d in list_dict_res:
         id = list(d.keys())[0]
         if id not in res.keys():
            res[id] = d[id]
            # will add stats for how many files this site has
            res[id]["file_num"] =1
            res[id]["components"]={}
            res[id]["matrix"]={}
         else:
            res[id]["file_num"] +=1
            res[id]["files"].update(d[id]["files"])
         
         # create a component indexing dict
         for f in d[id]["files"].keys():
            contents = d[id]["files"][f]["contents"]
            for c in contents:
               if c["component"] not in res[id]["components"].keys():
                  res[id]["components"][c["component"]] = [f]
               else:
                  res[id]["components"][c["component"]].append(f)
               
               if c["matrix"] not in res[id]["matrix"].keys():
                  res[id]["matrix"][c["matrix"]] = [f]
               else:
                  res[id]["matrix"][c["matrix"]].append(f)
                               
      return res